/*								STA/LTA Ratio Calculator
                                   Created 3/16/2001
 This program reads in a PSN Format 4 event file and then calculates 
 and displays the STA/LTA ratios. This information is also saved in a file
 named StaLta.TXT.

Updated: 4/07/2001		Added CRC16 test code

*/

#include <stdio.h>
#include <math.h>
#include <time.h>
#include "psn.h"
#include "crctbl.h"				// needed for CRC16 calculation

#define UINT	unsigned int
#define UINT64	unsigned _int64

/* The following variables hold the event file header information */
PSNType4 hdr;
EventInfo event;
PickInfo pick;
BYTE *notUsed;
char location[256], sensorType[256], comment[256];

/* Input file name */
char inFileName[256];

/* This array holds the sample data from the event file */
float *fBuff = 0;

/* STA/LTA calculation variables */
int staNum, ltaNum, staCnt, ltaCnt, staGood, ltaGood;
double staTime, ltaTime, staAvg, ltaAvg, sta, lta, maxRatio;

/* Function prototypes */
int parseGetInput(int, char *[]);
int readInFile(char *);
int checkFileCRC(FILE *, int);
UINT64 calcTime(DateTime *);
WORD calcFileCRC(FILE *, int);
void displayHeader(PSNType4 *);
void saveHeader(FILE *, PSNType4 *);
void calcStaLta();
void makeTimeString(char *, UINT64);
void calcCRC16(WORD *, BYTE *, int);
void printUsage();

/* Program starts here */
void main(int argc, char *argv[])
{
	if(!parseGetInput(argc, argv))		// get the user inputs
		return;
	
	if(readInFile(inFileName))  {		// try to read in the file
		displayHeader(&hdr);			// display header info if OK
		calcStaLta();					// calc STA/LTA Ratios
	}	
	
	if(fBuff)							// free the data buffer if allocated
		free(fBuff);
}

// Calculate the STA/LTA ratios
void calcStaLta()
{
	float *fptr, ftmp;
	int cnt, idx, ms, maxTimeIdx = 0;
	char str[80];
	double alpha;
	UINT64 tm;
	FILE *outfp;
		
	printf("\nCreating File: StaLta.txt\n");
	if(!(outfp = fopen("StaLta.txt", "w")))  {
		printf("Error creating file.\n");
		return;
	}
	saveHeader(outfp, &hdr);		// save the header info in the output file
	
	staNum = (int)(hdr.spsRate * staTime);	// calculate sta/lta times based
	ltaNum = (int)(hdr.spsRate * ltaTime);	// on sample rate
	
	staCnt = ltaCnt = staGood = ltaGood = 0;
	staAvg = ltaAvg = maxRatio = 0.0;
	
	cnt = hdr.sampleCount;			// number of samples to process
	printf("STA Sample Number: %d Time: %g  LTA Sample Number: %d Time: %g\n", 
			staNum, staTime, ltaNum, ltaTime);
	fprintf(outfp, "\nSTA Sample Number: %d Time: %g  LTA Sample Number: %d Time: %g\n\n", 
			staNum, staTime, ltaNum, ltaTime);
	
	fptr = fBuff;
	tm = calcTime(&hdr.startTime);	// get the start time in milliseconds
	
	idx = 0;
	while(cnt--)  {
		ftmp = *fptr++;
		if(ftmp < 0)				// rectify the data
			ftmp = -ftmp;
		staAvg += ftmp;				// average the STA and LTA data
		ltaAvg += ftmp;
		if(++ltaCnt >= ltaNum)  {	// calculate the LTA average
			lta = ltaAvg / (double)ltaCnt;
			if(!lta)
				lta = 0.001F;
			ltaAvg = 0.0F;
			ltaCnt = 0L;
			ltaGood = 1;
		} 
		if(++staCnt >= staNum)  {	// calculate the STA average
			sta = staAvg / (double)staCnt;
			staAvg = 0.0F;
			staCnt = 0L;
			staGood = 1;
		}
		++idx;						// next sample index
		if(!ltaGood || !staGood)	// wait for both too become good 
			continue;
		alpha = sta / lta;			// calculate the STA/LTA ratio
		if(alpha > maxRatio)  {		// keep track of max ratio
			maxRatio = alpha;
			maxTimeIdx = idx;		// used to calculate time
		}
		// make the current time string and save data to the output file
		ms = (int)(((double)idx * (1 / hdr.spsRate)) * 1000.0);
		makeTimeString(str, tm + ms);
		fprintf(outfp, "%s = %3.1f\n", str, alpha);
		staGood = 0;
	}
	// Save and display max ratio and time 
	ms = (int)(((double)maxTimeIdx * (1 / hdr.spsRate)) * 1000.0);
	makeTimeString(str, tm + ms);
	printf("Max STA/LTA Ratio = %3.2f at %s\n", maxRatio, str);
	fprintf(outfp, "Max STA/LTA Ratio = %3.2f at %s\n", maxRatio, str);
	fclose(outfp);
}

/* Read in the PSN Type 4 event file */
int readInFile(char *fname)
{
	short sData;
	long lData;
	float *fptr, fData;
	double dData;
	int cnt, dataSize, length;
	FILE *fp;
					
	if(!(fp = fopen(fname, "rb")))  {		// open the file
		printf("Can't open file.\n");
		return(0);
	}
	
	// read in the fixed header section
	if(fread(&hdr, 1, sizeof(PSNType4), fp) != sizeof(PSNType4))  {
		printf("Header read error.\n");
		fclose(fp);
		return(0);
	}
	
	if(strncmp(hdr.headerID, "PSNTYPE4", 8))  {
		printf("Invalid event file format.\n");
		fclose(fp);
		return(0);
	}
	
	if(!(hdr.flags & NO_CRC))  {		// see if valid CRC is present
		printf("Checking CRC16...");
		if(!hdr.sampleType)				// short, 2 bytes
			dataSize = sizeof(short);
		else if(hdr.sampleType == 1)	// long, 4 bytes
			dataSize = sizeof(long);
		else if(hdr.sampleType == 2)	// float, 4 bytes
			dataSize = sizeof(float);
		else 							// must be double, 4 bytes
			dataSize = sizeof(double);
		// calculate length to check
		length = sizeof(PSNType4) + hdr.varHdrLength + (dataSize * hdr.sampleCount);
		if(!checkFileCRC(fp, length))  {	// check the file
			printf("\nCRC Error - Press key to continue\n");
			getch();
		}
		else
			printf("Good CRC16\n");
	}
	else
		printf("No CRC16 to test.\n");	
	// if there is a variable length header section parse it
	if(hdr.varHdrLength)
		parseVarHdr(fp);
	
	cnt = hdr.sampleCount;
	fBuff = (float *)malloc(cnt * sizeof(float));	// allocate memory for data
	fptr = fBuff;
	
	// Now read in the sample data
	while(cnt--)  {
		if(!hdr.sampleType)  {				// short, 2 bytes
			if(fread(&sData, 1, 2, fp) != 2)  {
				printf("Data read error.\n");
				fclose(fp);
				return(0);
			}
			*fptr++ = sData;
		}
		else if(hdr.sampleType == 1)  {		// long, 4 bytes
			if(fread(&lData, 1, 4, fp) != 4)  {
				printf("Data read error.\n");
				fclose(fp);
				return(0);
			}
			*fptr++ = lData;
		}
		else if(hdr.sampleType == 2)  {		// float, 4 bytes
			if(fread(&fData, 1, 4, fp) != 4)  {
				printf("Data read error.\n");
				fclose(fp);
				return(0);
			}
			*fptr++ = fData;
		}
		else  {								// double, 8 bytes
			if(fread(&dData, 1, 8, fp) != 8)  {
				printf("Data read error.\n");
				fclose(fp);
				return(0);
			}
			*fptr++ = dData;
		}
	}
	fclose(fp);
	return(1);
}

/* Parse the variable length header part of the event file  */
int parseVarHdr(FILE *fp)
{
	VarHeader vhdr;
	int hdrLen = sizeof(VarHeader);
	char *tmp, string[2048];
		
	while(1)  {
		// First read in the data section header
		if(fread(&vhdr, 1, hdrLen, fp) != hdrLen)  {
			printf("VarHdr read error\n");
			return(0);
		}
		
		//  Sanity check, must be this number or error
		if(vhdr.checkNumber != 0x55)  {
			printf("VarHdr read error\n");
			return(0);
		}
		if(!vhdr.id)			// if id is zero we are done
			return(1);
//		printf("Hdr ID:%d\n", vhdr.id);
		
		// Get either sensor location, type or comment strings 
		if(vhdr.id == 1 || vhdr.id == 2 || vhdr.id == 3)  {
			if(fread(string, 1, vhdr.length, fp) != (UINT)vhdr.length)
				return(0);
			if(strlen(string) > 255)
				string[255] = 0;
			if(vhdr.id == 1)
				strcpy(location, string);
			else if(vhdr.id == 2)
				strcpy(sensorType, string);
			else if(vhdr.id == 3)
				strcpy(comment, string);
		}
		// Get event information. Note: there can be more then one of these in the file
		else if(vhdr.id == 4 && vhdr.length == sizeof(EventInfo))  {
			if(fread(&event, 1, vhdr.length, fp) != (UINT)vhdr.length)
				return(0);
		}
		// Get phase pick information. Note: there can be more then one of these in the file
		else if(vhdr.id == 5 && vhdr.length == sizeof(PickInfo))  {
			if(fread(&pick, 1, vhdr.length, fp) != (UINT)vhdr.length)
				return(0);
		}
		// Operations done to the data string
		else if(vhdr.id == 6 && vhdr.length)  {
			if(!(tmp = (char *)malloc(vhdr.length)))
				return(0);
			if(fread(tmp, 1, vhdr.length, fp) != (UINT)vhdr.length)
				return(0);
			free(tmp);
		}
		else  {		// process unknown types 
			if(!(notUsed = (BYTE *)malloc(hdrLen+vhdr.length)))
				return(0);
			memcpy(notUsed, &vhdr, hdrLen);
			if(fread(&notUsed[hdrLen], 1, vhdr.length, fp) != (UINT)vhdr.length)
				return(0);
			free(notUsed);
		}
	}
}

/* Displays the main header information on the screen */
void displayHeader(PSNType4 *p)
{
	char chr;
	
	DateTime *t = &p->startTime;
	printf("ID: %-8.8s\n    VarHdrLength:%ld StartTime:%02d:%02d:%02d %02d:%02d:%02d %ld Offset:%lg\n", 
		p->headerID, p->varHdrLength, t->month, t->day, t->year % 100, t->hour, t->minute, 
		t->second, t->nanoseconds, p->startTimeOffset);
	if(!p->timeRefStatus)
		chr = ' ';
	else
		chr = p->timeRefStatus;
	printf("    SPS:%lg Samples:%ld Flags:%x TimeRef:%-3.3s Status:%c", p->spsRate, 
		p->sampleCount, p->flags, p->timeRefType, chr);
	printf(" Type:%d Compress:%d\n", p->sampleType, p->sampleCompression);
	printf("    Orientation: %c Incident:%lg Azimuth:%lg Type:%d\n", p->compOrientation, 
		p->compIncident, p->compAzimuth, p->sensorType);	
	printf("    Sensor Lat:%lg Long:%lg Evl:%lg Name:%-6.6s Network:%-6.6s\n", 
		p->latitude, p->longitude, p->elevation, p->name, p->network);
	printf("    Sens:%lg MagCorr:%lg A/DBits:%d Min:%lg Max:%lg Mean:%lg\n", 
		p->sensitivity, p->magCorr, p->atodBits, p->minimum, p->maximum, p->mean);
}

/* Save the main header information to a file  */
void saveHeader(FILE *fp, PSNType4 *p)
{
	char chr;
	
	DateTime *t = &p->startTime;
	fprintf(fp, "ID: %-8.8s\n    VarHdrLength:%ld StartTime:%02d:%02d:%02d %02d:%02d:%02d %ld Offset:%lg\n", 
		p->headerID, p->varHdrLength, t->month, t->day, t->year % 100, t->hour, t->minute, 
		t->second, t->nanoseconds, p->startTimeOffset);
	if(!p->timeRefStatus)
		chr = ' ';
	else
		chr = p->timeRefStatus;
	fprintf(fp, "    SPS:%lg Samples:%ld Flags:%x TimeRef:%-3.3s Status:%c", p->spsRate, 
		p->sampleCount, p->flags, p->timeRefType, chr);
	fprintf(fp, " Type:%d Compress:%d\n", p->sampleType, p->sampleCompression);
	fprintf(fp, "    Orientation: %c Incident:%lg Azimuth:%lg Type:%d\n", p->compOrientation, 
		p->compIncident, p->compAzimuth, p->sensorType);	
	fprintf(fp, "    Sensor Lat:%lg Long:%lg Evl:%lg Name:%-6.6s Network:%-6.6s\n", 
		p->latitude, p->longitude, p->elevation, p->name, p->network);
	fprintf(fp, "    Sens:%lg MagCorr:%lg A/DBits:%d Min:%lg Max:%lg Mean:%lg\n", 
		p->sensitivity, p->magCorr, p->atodBits, p->minimum, p->maximum, p->mean);
}

/* Converts the start time to a 64 bit number in milliseconds past 1900. */
UINT64 calcTime(DateTime *dt)
{
	long tmptz = _timezone, ltime;
	int tmpdl = _daylight, ms;
	struct tm nt;
	UINT64 u64;
		
	_timezone = 0;
	_daylight = 0;
	memset(&nt, 0, sizeof(struct tm));
	nt.tm_sec = dt->second;
	nt.tm_min = dt->minute;
	nt.tm_hour = dt->hour;
	nt.tm_mon = dt->month - 1;
	nt.tm_mday = dt->day;
	nt.tm_year = dt->year - 1900;
	nt.tm_isdst = -1;
	ltime = mktime(&nt);
	ms = dt->nanoseconds / 1000000;
	u64 = ((UINT64)ltime * (UINT64)1000) + (UINT64)ms;	
	_daylight = tmpdl;
	_timezone = tmptz;
	return(u64);
}

/* Make a time string based on UINT64 tm */
void makeTimeString(char *to, UINT64 tm)
{
	long ltime, tmptz = _timezone;
	int tmpdl = _daylight, ms;
	struct tm *nt;
		
	ltime = (long)(tm / (UINT64)1000);
	ms = (int)(tm % (UINT64)1000);
	nt = localtime(&ltime);
	sprintf(to, "%02d:%02d:%02d.%03d", nt->tm_hour, nt->tm_min, nt->tm_sec, ms);
}

/* Parse user inputs or get data from user in not on the command line */
int parseGetInput(int argc, char *argv[])
{
	char tmp[80];
	
	if(argc <= 1)  {
		printf("File: ");
		gets(inFileName);
	}
	else  {
		if(argv[1][0] == '/' || argv[1][0] == '-')  {
			printUsage();
			return(0);
		}
		strcpy(inFileName, argv[1]);
	}
	if(argc <= 2)  {
		printf("STA Time: ");
		gets(tmp);
	}
	else
		strcpy(tmp, argv[2]);
	staTime = atof(tmp);
		
	if(argc <= 3)  {
		printf("LTA Time: ");
		gets(tmp);
	}
	else
		strcpy(tmp, argv[3]);
	ltaTime = atof(tmp);
	return(1);
}

/* Check the file for correct CRC16 */
int checkFileCRC(FILE *fp, int length)
{
	WORD crc, test = 0;
	long pos;
	crc = calcFileCRC(fp, length);			// calculate CRC16 up to the CRC data
	pos = ftell(fp);						// save the current file position
	fseek(fp, length, SEEK_SET);			// seek to the end of the file
	fread(&test, 1, sizeof(test), fp);		// read the CRC16 data
	fseek(fp, pos, SEEK_SET);				// restore file position
	if(crc == test)							// test CRC16
		return(1);							// return good
	return(0);								// return bad
}

/* Calculate the files CRC16 */
WORD calcFileCRC(FILE *fp, int length)
{
	BYTE buff[512];
	WORD crc = 0;
	int cnt, len;
	
	long pos = ftell(fp);					// save file position 
	fseek(fp, 0, SEEK_SET);					// seek to the beginning of the file
	while(length)  {						// now read in the file calculating
		if(length > 512)					// the CRC1 as we go
			len = 512;						// do chunks of 512 or less bytes
		else
			len = length;
		cnt = fread(buff, 1, len, fp);
		calcCRC16(&crc, buff, cnt);
		length -= len;
	}
	fseek(fp, pos, SEEK_SET);				// restore file position
	return(crc);							// return the CRC16
}

/* calculates the CRC16 for a chunk of data */
void calcCRC16(WORD *crc, BYTE *cp, int cnt)
{
	while(cnt--)
		*crc = (*crc << B) ^ crctab[(*crc>>(W - B)) ^ *cp++];
}

/* Print to the screen how to use the program */
void printUsage()
{
	printf("Usage: StaLta [Input File] [STA Time] [LTA Time]\n");
	printf("       Input File must be a PSN Type 4 formatted event file.\n");
	printf("       STA and LTA times are in seconds.\n");
	printf("       Example: StaLta 010210a.lc1 .5 20.0\n");
}
